# SPDX-License-Identifier: Apache-2.0
# Copyright © 2021 Wrangle Ltd

import datetime
import typing

import attr

from wrgl.isoformat import fromisoformat
from wrgl.serialize import field_transformer


@attr.s(auto_attribs=True, field_transformer=field_transformer(globals()))
class CommitResult(object):
    """Payload of a successful commit

    :ivar str sum: 16-bit checksum of commit presented as hex string
    :ivar str table: 16-bit checksum of underlying table presented as hex string
    """
    sum: str
    table: str


@attr.s(auto_attribs=True, field_transformer=field_transformer(globals()))
class Table(object):
    """A table represents the underlying CSV of a :class:`Commit`

    :ivar str sum: 16-bit checksum of table presented as hex string
    :ivar list[str] columns: list of column names
    :ivar list[int] pk: indices of primary key columns
    :ivar int rows_count: number of rows
    """
    sum: str
    columns: typing.List[str]
    pk: typing.List[int]
    rows_count: int

    @property
    def primary_key(self) -> typing.List[str]:
        """Returns primary-key columns of the first commit

        :rtype: list[str]
        """
        if self.pk is None:
            return []
        return [self.columns[idx] for idx in self.pk]


@attr.s(auto_attribs=True, field_transformer=field_transformer(globals()))
class Commit(object):
    """A commit represents an immutable snapshot of a CSV

    :ivar str sum: 16-bit checksum of commit presented as hex string
    :ivar str author_name: name of commit author
    :ivar str author_email: email of commit author
    :ivar str message: commit message
    :ivar Table table: the underlying table
    :ivar datetime.datetime time: commit time
    :ivar list[str] parents: checksums of parent commits
    :ivar dict[str, Commit] parent_commits: mapping between parent checksums and commits.
        Only present if the commit was returned by :func:`Repository.get_commit_tree`
    """
    sum: str
    author_name: str
    author_email: str
    message: str
    table: Table
    time: datetime.datetime = attr.ib(
        converter=attr.converters.optional(fromisoformat)
    )
    parents: typing.List[str]
    parent_commits: "typing.Dict[str, Commit]"


@attr.s(auto_attribs=True, field_transformer=field_transformer(globals()))
class CommitTree(object):
    """Payload returned by :func:`Repository.get_commit_tree`

    :ivar str sum: 16-bit checksum of root commit presented as hex string
    :ivar Commit root: the root commit
    """
    sum: str
    root: Commit



# SPDX-License-Identifier: Apache-2.0
# Copyright © 2021 Wrangle Ltd

from typing import Iterator, List
import tempfile
import gzip
import shutil
import typing
import requests
import csv
import io
from requests_toolbelt.multipart.encoder import MultipartEncoder

from wrgl import diffreader
from wrgl.commit import Commit, CommitResult, Table, CommitTree
from wrgl.diff import DiffResult
from wrgl.serialize import json_dumps, json_loads
from wrgl.config import Config


class Repository(object):
    """Represents the HTTP API that wraps a hosted Wrgl repository
    """

    def __init__(self, endpoint: str, id_token: str = None) -> None:
        """
        :param str endpoint: the endpoint of the HTTP API
        :param str id_token: optional, the access token for this API.
            If it is a valid token then you don't need to run :func:`Repository.authenticate`.
        """
        self.endpoint = endpoint.rstrip("/")
        self._is_hub_repo = self.endpoint.startswith("https://hub.wrgl.co/")
        self._id_token = id_token

    def authenticate(self, email: str, password: str) -> str:
        """Authenticates and returns the access token.

        If successful, access token will be saved for future usage.

        :param str email: user's email
        :param str password: user's password

        :rtype: str
        """
        if self._is_hub_repo:
            endpoint = "https://hub.wrgl.co/api/authenticate/"
        else:
            endpoint = self.endpoint + "/authenticate/"
        r = requests.post(endpoint, json={
            'email': email,
            'password': password
        })
        r.raise_for_status()
        self._id_token = r.json()['idToken']
        return self._id_token

    def _headers(self, extra_headers=dict()):
        headers = list(extra_headers.items())
        if self._id_token is not None:
            headers.append(('Authorization', 'Bearer %s' % self._id_token))
        return dict(headers)

    def _get(self, path, params=None):
        if params is not None:
            params = {k: v for k, v in params.items() if v}
        r = requests.get(
            self.endpoint+path, params=params, headers=self._headers(),
        )
        r.raise_for_status()
        return r

    def _put_json(self, path, obj):
        r = requests.put(
            self.endpoint+path,
            data=json_dumps(obj),
            headers=self._headers({
                'Content-Type': 'application/json'
            }),
        )
        r.raise_for_status()
        return r

    def _post_json(self, path, obj):
        r = requests.post(
            self.endpoint+path,
            data=json_dumps(obj),
            headers=self._headers({
                'Content-Type': 'application/json'
            }),
        )
        r.raise_for_status()
        return r

    def get_refs(self) -> dict:
        """Get references as a mapping of reference name and commit checksum

        :rtype: dict
        """
        r = self._get("/refs/")
        obj = r.json()
        return obj["refs"]

    def get_branch(self, branch: str) -> Commit:
        """Get the head commit of a branch

        :param str branch: the name of the branch

        :rtype: Commit
        """
        r = self._get("/refs/heads/%s/" % branch)
        return json_loads(r.content, Commit)

    def commit(self, branch: str, message: str, file: typing.BinaryIO, primary_key: typing.List[str]) -> CommitResult:
        """Creates a new commit

        :param str branch: name of the branch to commit under
        :param str message: commit message
        :param typing.BinaryIO file: the CSV file to commit
        :param list[str] primary_key: list of column names that make up the primary key

        :rtype: CommitResult
        """
        with tempfile.TemporaryFile() as fp:
            with gzip.open(fp, 'w') as gzf:
                shutil.copyfileobj(file, gzf)
            fp.seek(0)
            m = MultipartEncoder(
                fields={
                    'branch': branch,
                    'message': message,
                    'primaryKey': ','.join(primary_key),
                    'file': ('data.csv.gz', fp, 'text/csv')
                }
            )
            r = requests.post(
                self.endpoint+"/commits/",
                data=m,
                headers=self._headers({'Content-Type': m.content_type}),
            )
        r.raise_for_status()
        return json_loads(r.content, CommitResult)

    def get_commit_tree(self, head: str, max_depth: int) -> CommitTree:
        """Gets commit tree

        :param str head: name of the root commit, could either be reference name or commit checksum.
        :param int max_depth: maximum depth of commit tree to fetch

        :rtype: CommitTree
        """
        r = self._get(
            "/commits/", params={'head': head, 'maxDepth': max_depth}
        )
        return json_loads(r.content, CommitTree)

    def get_commit(self, commit_sum: str) -> Commit:
        """Get commit with the given checksum

        :param str commit_sum: commit checksum

        :rtype: Commit
        """
        r = self._get("/commits/%s/" % commit_sum)
        return json_loads(r.content, Commit)

    def get_table(self, table_sum: str) -> Table:
        """Get table with the given checksum

        :param str table_sum: table checksum

        :rtype: Table
        """
        r = self._get("/tables/%s/" % table_sum)
        return json_loads(r.content, Table)

    def get_blocks(self, commit: str, start: int = None, end: int = None, with_column_names: bool = True) -> Iterator[List[str]]:
        """Fetchs blocks as concatenated rows. Each row as a list of strings.

        Calling this with default `start`, `end`, and `with_column_names` will return the entire table.

        :param str commit: either commit checksum or reference e.g. "heads/main"
        :param int start: index of the first block to fetch. Defaults to 0.
        :param int end: index of the last block to fetch. If not set, fetch til the end.
        :param bool with_column_names: prepend column names to the resulting CSV, which in effect producing a CSV with header.

        :rtype: typing.Iterator[list[str]]
        """
        r = self._get(
            "/blocks/",
            params={
                'head': commit,
                'start': start,
                'end': end,
                'columns': 'true' if with_column_names else 'false'
            }
        )
        for row in csv.reader(io.StringIO(r.text), dialect='unix'):
            yield row

    def get_table_blocks(self, table_sum: str, start: int = None, end: int = None, with_column_names: bool = True) -> Iterator[List[str]]:
        """Fetchs blocks with table checksum.

        Calling this with default `start`, `end`, and `with_column_names` will return the entire table.

        :param str table_sum: table checksum
        :param int start: index of the first block to fetch. Defaults to 0.
        :param int end: index of the last block to fetch. If not set, fetch til the end.
        :param bool with_column_names: prepend column names to the resulting CSV, which in effect producing a CSV with header.

        :rtype: typing.Iterator[list[str]]
        """
        r = self._get(
            "/tables/%s/blocks/" % table_sum,
            params={
                'start': start,
                'end': end,
                'columns': 'true' if with_column_names else 'false'
            }
        )
        for row in csv.reader(io.StringIO(r.text), dialect='unix'):
            yield row

    def get_rows(self, commit: str, offsets: List[int]) -> Iterator[List[str]]:
        """Get rows at certain offsets. Each row will be returned as a list of strings.

        This is usually used in tandem with row offsets from :class:`DiffResult` to fetch changed rows.

        :param str commit: either commit checksum or reference e.g. "heads/main"
        :param list[int] offsets: the offsets of the rows to fetch

        :rtype: typing.Iterator[list[str]]
        """
        r = self._get(
            "/rows/",
            params={
                'head': commit,
                'offsets': ','.join([str(v) for v in offsets])
            }
        )
        for row in csv.reader(io.StringIO(r.text), dialect='unix'):
            yield row

    def get_table_rows(self, table_sum: str, offsets: List[int]) -> Iterator[List[str]]:
        """Get rows at certain offsets with table checksum.

        This is usually used in tandem with row offsets from :class:`DiffResult` to fetch changed rows.

        :param str table_sum: table checksum
        :param list[int] offsets: the offsets of the rows to fetch

        :rtype: typing.Iterator[list[str]]
        """
        r = self._get(
            "/tables/%s/rows/" % table_sum,
            params={
                'offsets': ','.join([str(v) for v in offsets])
            }
        )
        for row in csv.reader(io.StringIO(r.text), dialect='unix'):
            yield row

    def diff(self, sum1: str, sum2: str) -> DiffResult:
        """Compares two commits and returns their differences.

        :param str sum1: checksum of the first commit
        :param str sum2: checksum of the second commit

        :rtype: DiffResult
        """
        r = self._get(
            "/diff/%s/%s/" % (sum1, sum2)
        )
        return json_loads(r.content, DiffResult)

    def diff_reader(self, sum1: str, sum2: str, fetch_size: int = 100) -> diffreader.DiffReader:
        """Compares two commits and interpret their differences.

        This method is higher level than :func:`Repository.diff` and should
        be preferred for 99% of use cases.

        :param str sum1: checksum of the first commit
        :param str sum2: checksum of the second commit

        :rtype: DiffReader
        """
        return diffreader.DiffReader(self, sum1, sum2, fetch_size)



from unittest import TestCase
import datetime

from wrgl.isoformat import fromisoformat

class FromISOFormatTestCase(TestCase):
    def test_parse(self):
        self.assertEqual(
            fromisoformat('2021-11-17T01:22:53Z'),
            datetime.datetime(2021, 11, 17, 1, 22, 53, tzinfo=datetime.timezone.utc)
        )
        self.assertEqual(
            fromisoformat('2021-11-17T01:22:53+07:00'),
            datetime.datetime(2021, 11, 17, 1, 22, 53, tzinfo=datetime.timezone(datetime.timedelta(hours=7)))
        )


# SPDX-License-Identifier: Apache-2.0
# Copyright © 2021 Wrangle Ltd

from datetime import datetime
import typing
import platform
import _csv
from unittest import TestCase
from contextlib import contextmanager
import io
import tempfile
import pathlib
import tarfile
import socket
import re
import time
import os
import subprocess
import csv

import requests

from wrgl.config import Config, Receive, User
from wrgl.diffreader import ColumnChanges
from wrgl.repository import Repository


def read_version():
    pat = re.compile(r'^version = (\d+\.\d+\.\d+)')
    with open(pathlib.Path(__file__).parent.parent / 'setup.cfg', 'r') as f:
        for line in f:
            m = pat.match(line)
            if m is not None:
                return m.group(1)
    raise ValueError('cannot read version from file setup.cfg')


def download_wrgl(version):
    ver_dir = pathlib.Path(__file__).parent / '__testcache__' / version
    OS = platform.system().lower()
    if not ver_dir.exists():
        ver_dir.mkdir(parents=True)
        for url in [
                "https://github.com/wrgl/wrgl/releases/download/v%s/wrgl-%s-amd64.tar.gz" % (
                    version, OS
                ), "https://github.com/wrgl/wrgl/releases/download/v%s/wrgld-%s-amd64.tar.gz" % (
                    version, OS
                )]:
            with requests.get(url, stream=True) as r:
                with tarfile.open(mode='r:gz', fileobj=r.raw) as tar:
                    tar.extractall(ver_dir)
    return ver_dir / ('wrgl-%s-amd64' % OS) / 'bin' / 'wrgl', ver_dir / ('wrgld-%s-amd64' % OS) / 'bin' / 'wrgld'


def get_open_port():
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.bind(("", 0))
    s.listen(1)
    port = s.getsockname()[1]
    s.close()
    return str(port)


class RepositoryTestCase(TestCase):
    maxDiff = None

    @classmethod
    def setUpClass(cls):
        cls.version = read_version()
        cls.wrgl_bin, cls.wrgld_bin = download_wrgl(cls.version)

    def wrgl(self, args, print_result=False):
        try:
            result = subprocess.run(
                [str(self.wrgl_bin)]+list(args) +
                ["--wrgl-dir", self.repo_dir.name],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                check=True)
        except subprocess.CalledProcessError as e:
            raise Exception("%s: %s\n  stdout: %s\n  stderr: %s" %
                            (type(e).__name__, e.args, e.stdout, e.stderr))
        if print_result:
            print("%s\n  stdout: %s\n  stderr: %s" %
                  (" ".join(result.args), result.stdout, result.stderr))
        return result

    @contextmanager
    def run_wrgld(self) -> typing.Iterator[str]:
        port = get_open_port()
        proc = subprocess.Popen(
            [str(self.wrgld_bin), self.repo_dir.name, "-p", port], stdout=subprocess.PIPE)
        time.sleep(1)
        yield "http://localhost:%s" % port
        proc.stdout.close()
        proc.terminate()
        proc.wait()

    def setUp(self):
        super().setUp()
        self.repo_dir = tempfile.TemporaryDirectory()
        self.wrgl(["init"])
        self.email = "johndoe@domain.com"
        self.name = "John Doe"
        self.password = "password"
        self.wrgl(
            ["config", "set", "receive.denyNonFastForwards", "true"]
        )
        self.wrgl(
            ["config", "set", "user.email", self.email]
        )
        self.wrgl(
            ["config", "set", "user.name", self.name]
        )
        self.wrgl(
            ["auth", "add-user", self.email, "--name",
                self.name, "--password", self.password]
        )
        self.wrgl(
            ["auth", "add-scope", self.email, "--all"]
        )
        self.wrgl(["config", "set", "auth.type", "legacy"])
        self.wrgl(["config", "get", "auth.type"], print_result=True)

    def tearDown(self) -> None:
        self.repo_dir.cleanup()
        return super().tearDown()

    @contextmanager
    def commit(self, branch: str, message: str, primary_key: typing.List[str]) -> typing.Iterator["_csv._writer"]:
        with tempfile.NamedTemporaryFile(mode='w', delete=False) as f:
            writer = csv.writer(f)
            yield writer
        try:
            self.wrgl(
                ["commit", branch, f.name, message,
                    "-p", ",".join(primary_key)]
            )
        finally:
            os.remove(f.name)

    def test_commit(self):
        with self.commit("main", "initial commit", ["a"]) as writer:
            writer.writerow(['a', 'b', 'c'])
            writer.writerow(['1', 'q', 'w'])
            writer.writerow(['2', 'a', 's'])

        with self.run_wrgld() as url:
            repo = Repository(url)
            repo.authenticate(self.email, self.password)

            refs = repo.get_refs()
            self.assertEqual(len(refs), 1)
            commit1 = repo.get_commit(refs['heads/main'])
            self.assertIsInstance(commit1.time, datetime)
            commit2 = repo.get_branch('main')
            print(commit2)
            self.assertEqual(commit1, commit2)

            b = io.StringIO()
            writer = csv.writer(b)
            writer.writerow(['a', 'b', 'c'])
            writer.writerow(['1', 'e', 'w'])
            writer.writerow(['2', 'c', 's'])
            cr = repo.commit(
                branch='main',
                message='second commit',
                file=io.BytesIO(b.getvalue().encode('utf8')),
                primary_key=['a']
            )
            self.assertIsNotNone(cr.sum)
            self.assertIsNotNone(cr.table)

            com_tree = repo.get_commit_tree(cr.sum, 2)
            self.assertIsNotNone(
                com_tree.root.parent_commits[commit1.sum].table
            )

            tbl = repo.get_table(cr.table)
            self.assertEqual(tbl.columns, ['a', 'b', 'c'])
            self.assertEqual(list(repo.get_blocks(cr.sum)), [
                ['a', 'b', 'c'], ['1', 'e', 'w'], ['2', 'c', 's']
            ])
            self.assertEqual(list(repo.get_table_blocks(cr.table)), [
                ['a', 'b', 'c'], ['1', 'e', 'w'], ['2', 'c', 's']
            ])
            self.assertEqual(list(repo.get_rows(cr.sum, [0])), [
                ['1', 'e', 'w']
            ])
            self.assertEqual(list(repo.get_table_rows(cr.table, [0])), [
                ['1', 'e', 'w']
            ])
            dr = repo.diff(cr.sum, commit1.sum)
            self.assertTrue(dr.primary_key == dr.old_primary_key)
            self.assertGreater(len(dr.row_diff), 0)

    def test_diff_reader(self):
        with self.commit("main", "initial commit", ["a"]) as writer:
            writer.writerow(['a', 'b', 'c', 'd'])
            writer.writerow(['1', 'q', 'w', 'e'])
            writer.writerow(['2', 'a', 's', 'd'])
            writer.writerow(['3', 'z', 'x', 'c'])

        with self.commit("main", "second commit", ["a"]) as writer:
            writer.writerow(['a', 'b', 'c', 'e'])
            writer.writerow(['1', 'q', 'u', 'r'])
            writer.writerow(['2', 'a', 's', 'f'])
            writer.writerow(['4', 'y', 'u', 'i'])

        with self.run_wrgld() as url:
            repo = Repository(url)
            token = repo.authenticate(self.email, self.password)

        with self.run_wrgld() as url:
            repo = Repository(url, token)
            com_tree = repo.get_commit_tree("heads/main", 2)
            dr = repo.diff_reader(com_tree.sum, com_tree.root.parents[0])
            self.assertEqual(dr.column_changes, ColumnChanges(
                new_values=['a', 'b', 'c', 'e'],
                old_values=['a', 'b', 'c', 'd'],
                unchanged=set(['a', 'b', 'c']),
                added=set(['e']),
                removed=set(['d'])
            ))
            self.assertEqual(dr.pk_changes, ColumnChanges(
                new_values=['a'],
                old_values=['a'],
                unchanged=set(['a']),
                added=set([]),
                removed=set([])
            ))
            self.assertEqual(len(dr.added_rows), 1)
            self.assertEqual(dr.added_rows.columns, ['a', 'b', 'c', 'e'])
            self.assertEqual(dr.added_rows.primary_key, ['a'])
            self.assertEqual(list(dr.added_rows), [
                ['4', 'y', 'u', 'i']
            ])
            self.assertEqual(len(dr.removed_rows), 1)
            self.assertEqual(dr.removed_rows.columns, ['a', 'b', 'c', 'd'])
            self.assertEqual(dr.removed_rows.primary_key, ['a'])
            self.assertEqual(list(dr.removed_rows), [
                ['3', 'z', 'x', 'c']
            ])
            self.assertEqual(len(dr.modified_rows), 2)
            self.assertEqual(
                dr.modified_rows.columns, ['a', 'b', 'c', 'd', 'e']
            )
            self.assertEqual(dr.modified_rows.primary_key, ['a'])
            self.assertEqual(list(dr.modified_rows), [
                [('1', '1'), ('q', 'q'), ('u', 'w'), (None, 'e'), ('r', None)],
                [('2', '2'), ('a', 'a'), ('s', 's'), (None, 'd'), ('f', None)]
            ])
            self.assertIsNotNone(dr.data_profile)
            self.assertEqual(dr.data_profile.old_rows_count, 3)
            self.assertEqual(dr.data_profile.new_rows_count, 3)
            self.assertEqual(len(dr.data_profile.columns), 5)

    def test_diff_reader_no_changes(self):
        with self.commit("main", "initial commit", ["a"]) as writer:
            writer.writerow(['a', 'b', 'c', 'd'])
            writer.writerow(['1', 'q', 'w', 'e'])
            writer.writerow(['2', 'a', 's', 'd'])
            writer.writerow(['3', 'z', 'x', 'c'])

        with self.commit("main", "second commit", ["a"]) as writer:
            writer.writerow(['a', 'b', 'c', 'd'])
            writer.writerow(['1', 'q', 'w', 'e'])
            writer.writerow(['2', 'a', 's', 'd'])
            writer.writerow(['3', 'z', 'x', 'c'])

        with self.run_wrgld() as url:
            repo = Repository(url)
            repo.authenticate(self.email, self.password)
            com_tree = repo.get_commit_tree("heads/main", 2)
            dr = repo.diff_reader(com_tree.sum, com_tree.root.parents[0])
            self.assertEqual(dr.column_changes, ColumnChanges(
                new_values=['a', 'b', 'c', 'd'],
                old_values=['a', 'b', 'c', 'd'],
                unchanged=set(['a', 'b', 'c', 'd']),
                added=set([]),
                removed=set([])
            ))
            self.assertEqual(dr.pk_changes, ColumnChanges(
                new_values=['a'],
                old_values=['a'],
                unchanged=set(['a']),
                added=set([]),
                removed=set([])
            ))
            self.assertIsNone(dr.added_rows)
            self.assertIsNone(dr.removed_rows)
            self.assertIsNone(dr.modified_rows)



# SPDX-License-Identifier: Apache-2.0
# Copyright © 2021 Wrangle Ltd

from wrgl.config import Config, User, Remote, Branch, Receive, Auth, Pack
from wrgl.commit import Commit, CommitResult, CommitTree, Table
from wrgl.diff import DiffResult, RowDiff
from wrgl.repository import Repository

__all__ = [
    "Config", "User", "Remote", "Branch", "Receive", "Auth", "Pack",
    "Commit", "CommitResult", "CommitTree", "Table",
    "DiffResult", "RowDiff",
    "Repository",
]



# SPDX-License-Identifier: Apache-2.0
# Copyright © 2021 Wrangle Ltd

import attr
import typing
import math

from wrgl.commit import Table
from wrgl.serialize import none_field_transformer


def longest_increasing_list(values: typing.List[int]) -> typing.List[int]:
    """Returns indices of longest increasing values

    :param list[int] values: arbitrary integers

    :rtype: list[int]
    """
    @attr.s(auto_attribs=True)
    class Node(object):
        ind: int
        len: int
        prev: "Node"

    nodes: typing.Dict[int, Node] = dict()
    root: Node = None
    for i, v in enumerate(values):
        prev: Node = None
        for j in range(v-1, -1, -1):
            if j in nodes and (prev is None or prev.len < nodes[j].len):
                prev = nodes[j]
            if prev is not None and j < prev.len:
                break
        nodes[v] = Node(ind=i, len=1, prev=prev)
        if prev is not None:
            nodes[v].len = prev.len + 1
        if root is None or root.len < nodes[v].len or (root.len == nodes[v].len and v == i):
            root = nodes[v]
    results = []
    while root is not None:
        results.insert(0, root.ind)
        root = root.prev
    return results


@attr.s(auto_attribs=True)
class MoveOp(object):
    old_ind: int
    new_ind: int


def moveOps(sl: typing.List[int]) -> typing.List[MoveOp]:
    """moveOps returns move operations that changed order of array indices
    """
    anchor_indices = longest_increasing_list(sl)
    for i in anchor_indices:
        sl[i] = -1
    ops: typing.List[MoveOp] = []
    for i, v in enumerate(sl):
        if v == -1:
            continue
        ops.append(MoveOp(old_ind=v, new_ind=i))
    return ops


@attr.s(auto_attribs=True, field_transformer=none_field_transformer)
class Move(object):
    after: int
    before: int


@attr.s(auto_attribs=True, field_transformer=none_field_transformer)
class Column(object):
    name: str = attr.ib()
    base_idx: int = attr.ib()
    base_pk: bool = attr.ib()
    added: typing.Set[int] = attr.ib(factory=set)
    removed: typing.Set[int] = attr.ib(factory=set)
    moved: typing.Dict[int, Move] = attr.ib(factory=dict)
    layer_idx: typing.Dict[int, int] = attr.ib(factory=dict)
    layer_pk: typing.Set[int] = attr.ib(factory=set)

    @property
    def is_added(self) -> bool:
        return len(self.added) > 0

    @property
    def is_removed(self) -> bool:
        return len(self.removed) > 0

    @property
    def is_moved(self) -> bool:
        return len(self.moved) > 0


class ColDiff(object):
    """Keeps track of how column composition and order change between a base version and one or more versions
    """

    columns: typing.List[Column]
    name_map: typing.Dict

    def __init__(self, base: Table, *layers: Table) -> None:
        self.columns = []
        self.name_map = dict()
        for layer in layers:
            self.insert_columns(layer.columns)
        self.insert_columns(base.columns)
        for i, layer in enumerate(layers):
            self.assign_column_attrs(base.columns, i, layer.columns)
        self.hoist_pk_to_start({
            v: i for i, v in enumerate(layers[0].primary_key)
        })
        self.assign_index(base, *layers)

    def hoist_pk_to_start(self, pk: typing.Dict[str, int]) -> None:
        self.columns = sorted(
            self.columns, key=lambda col: pk.get(col.name, math.inf))
        self.name_map = {
            s: i for i, s in enumerate([
                col.name for col in self.columns
            ])
        }

    def assign_index(self, base: Table, *layers: Table) -> None:
        for i, name in enumerate(base.columns):
            self.columns[self.name_map[name]].base_idx = i
        for name in base.primary_key:
            self.columns[self.name_map[name]].base_pk = True
        for i, layer in enumerate(layers):
            for j, name in enumerate(layer.columns):
                self.columns[self.name_map[name]].layer_idx[i] = j
            for name in layer.primary_key:
                self.columns[self.name_map[name]].layer_pk.add(i)

    def insert_columns(self, column_names: typing.List[str]) -> None:
        offset = -1
        insert_map: typing.Dict[int, typing.List[Column]] = dict()
        for name in column_names:
            if name in self.name_map:
                offset = self.name_map[name]
                continue
            insert_map.setdefault(offset, []).append(Column(name=name))
        if len(insert_map) == 0:
            return
        inserts = sorted(
            list(insert_map.items()),
            key=lambda pair: -pair[0]
        )
        for insert_off, cols in inserts:
            self.columns[insert_off+1:insert_off+1] = cols
        self.name_map = {
            s: i for i, s in enumerate([
                col.name for col in self.columns
            ])
        }

    def assign_column_attrs(self, base_cols: typing.List[str], layer_idx: int, layer_cols: typing.List[str]) -> None:
        base_set = set(base_cols)
        layer_set = set(layer_cols)
        for name in layer_cols:
            if name not in base_set:
                self.columns[self.name_map[name]].added.add(layer_idx)
        for name in base_cols:
            if name not in layer_set:
                self.columns[self.name_map[name]].removed.add(layer_idx)
        self.assign_column_moved(base_cols, layer_idx, layer_set)

    def assign_column_moved(self, base_cols: typing.List[str], layer_idx: int, layer_set: typing.Set[str]) -> None:
        common_cols = [
            name for name in base_cols
            if name in layer_set
        ]
        common_map = {
            name: i for i, name in enumerate(common_cols)
        }
        old_indices = []
        new_indices = []
        for i, col in enumerate(self.columns):
            if col.name in common_map:
                new_indices.append(i)
                old_indices.append(common_map[col.name])
        ops = moveOps(old_indices)
        non_anchor = set([v.old_ind for v in ops])
        for op in ops:
            new_ind = new_indices[op.new_ind]
            after = None
            for i in range(op.old_ind-1, -1, -1):
                if i in non_anchor:
                    continue
                after = common_cols[i]
                if after in self.name_map:
                    break
            if after is not None:
                self.columns[new_ind].moved[layer_idx] = Move(
                    after=self.name_map[after]
                )
                continue

            before = None
            for i in range(op.old_ind, len(common_cols)):
                if i in non_anchor:
                    continue
                before = common_cols[i]
                if before in self.name_map:
                    break
            if before is not None:
                self.columns[new_ind].moved[layer_idx] = Move(
                    before=self.name_map[before]
                )

    def rearrange_row(self, layer: int, row: typing.List[str]) -> typing.List[str]:
        return [
            None if col.layer_idx is None or layer not in col.layer_idx
            else row[col.layer_idx[layer]]
            for col in self.columns
        ]

    def rearrange_base_row(self, row: typing.List[str]) -> typing.List[str]:
        return [
            None if col.base_idx is None else row[col.base_idx]
            for col in self.columns
        ]

    def combine_rows(self, layer: int, new_row: typing.List[str], old_row: typing.List[str]) -> typing.List[typing.Tuple[str or None, str or None]]:
        return [
            (
                None if layer in col.removed else new_row[col.layer_idx[layer]],
                None if layer in col.added else old_row[col.base_idx]
            )
            for col in self.columns
        ]

    def no_column_changes(self) -> bool:
        for col in self.columns:
            if len(col.added) > 0 or len(col.removed) > 0 or len(col.moved) > 0:
                return False
        return True



# SPDX-License-Identifier: Apache-2.0
# Copyright © 2021 Wrangle Ltdimport typing

import itertools
import typing
import attr

from wrgl import repository
from wrgl.commit import Table
from wrgl.diff import DiffResult, TableProfileDiff
from wrgl.coldiff import ColDiff


class RowIterator(object):
    """Iterates over rows with specified offsets of a table.

    Each row is returned as a list of strings.

    :var list[str] columns: column names
    :var list[str] primary_key: primary key
    """
    _repo: "repository.Repository"
    _tbl_sum: str
    _offsets: typing.List[int]
    _off: int
    _fetch_size: int

    columns: typing.List[str]
    primary_key: typing.List[str]

    def __init__(
            self,
            repo: "repository.Repository",
            tbl_sum: str,
            columns: typing.List[str],
            primary_key: typing.List[str],
            fetch_size: int = 100
    ) -> None:
        """
        :param Repository repo: the repo handle
        :param str tbl_sum: checksum of the table the rows belong to
        :param list[str] columns: column names
        :param list[str] primary_key: primary key
        :param int fetch_size: number of rows to fetch for each batch
        """
        self._repo = repo
        self._tbl_sum = tbl_sum
        self._offsets = []
        self._fetch_size = fetch_size
        self.columns = columns
        self.primary_key = primary_key

    def add_offset(self, offset: int) -> None:
        """Add a single row offset

        :param int offset: row offset
        """
        self._offsets.append(offset)

    def __len__(self):
        return len(self._offsets)

    def __iter__(self):
        self._off = 0
        self._batch = (i for i in [])
        return self

    def __next__(self) -> typing.List[str]:
        try:
            return next(self._batch)
        except StopIteration:
            if self._off >= len(self):
                raise StopIteration()
            self._batch = self._repo.get_table_rows(
                self._tbl_sum,
                self._offsets[self._off:self._off + self._fetch_size]
            )
            self._off += self._fetch_size
            return next(self._batch)


class ModifiedRowIterator(object):
    """Iterates over row pairs with specifies offsets from a pair of tables

    Each row is returned as a list of tuple of two values: `(newer_value, older_value)`.
    If either cell is missing (because the column is missing in one of the tables) then
    one of the values is None.

    :var list[str] columns: column names
    :var list[str] primary_key: primary key
    """
    _repo: "repository.Repository"
    _tbl_sum1: str
    _tbl_sum2: str
    _cd: ColDiff
    _fetch_size: int
    _offsets: typing.List[typing.Tuple[int, int]]
    _off: int

    columns: typing.List[str]
    primary_key: typing.List[str]

    def __init__(
            self,
            repo: "repository.Repository",
            tbl_sum1: str,
            tbl_sum2: str,
            cd: ColDiff,
            columns: typing.List[str],
            primary_key: typing.List[str],
            fetch_size: int = 100
    ) -> None:
        """
        :param Repository repo: the repo handle
        :param str tbl_sum1: checksum of the newer table
        :param str tbl_sum2: checksum of the older table
        :param ColDiff cd: column differences
        :param list[str] columns: column names
        :param list[str] primary_key: primary key
        :param int fetch_size: number of rows to fetch for each batch
        """
        self._repo = repo
        self._tbl_sum1 = tbl_sum1
        self._tbl_sum2 = tbl_sum2
        self._cd = cd
        self._fetch_size = fetch_size
        self._offsets = []
        self._off = 0
        self.columns = columns
        self.primary_key = primary_key

    def add_offset(self, offset1: int, offset2: int) -> None:
        """Add a single row offset

        :param int offset1: row offset for the newer table
        :param int offset2: row offset for the older table
        """
        self._offsets.append((offset1, offset2))

    def __len__(self):
        return len(self._offsets)

    def __iter__(self):
        self._off = 0
        self._batch = (i for i in [])
        return self

    def __next__(self) -> typing.List[str]:
        try:
            row1, row2 = next(self._batch)
        except StopIteration:
            if self._off >= len(self):
                raise StopIteration()
            offsets = self._offsets[self._off:self._off + self._fetch_size]
            self._batch = itertools.zip_longest(
                self._repo.get_table_rows(
                    self._tbl_sum1, [i for i, _ in offsets]
                ),
                self._repo.get_table_rows(
                    self._tbl_sum2, [i for _, i in offsets]
                )
            )
            self._off += self._fetch_size
            row1, row2 = next(self._batch)
        return self._cd.combine_rows(0, row1, row2)


@attr.s(auto_attribs=True)
class ColumnChanges(object):
    """Represents changes in column names.

    :var list[str] new_values: new columns
    :var list[str] old_values: old columns
    :var set[str] unchanged: columns that are present in both versions
    :var set[str] added: columns that appear in newer version but not in older version
    :var set[str] removed: columns that appear in older version but not in newer version
    """
    new_values: typing.List[str]
    old_values: typing.List[str]
    unchanged: typing.Set[str]
    added: typing.Set[str]
    removed: typing.Set[str]

    @classmethod
    def from_new_old_columns(cls, new_cols: typing.List[str], old_cols: typing.List[str]) -> "ColumnChanges":
        """Creates a new instance by comparing two column lists

        :param list[str] new_cols: newer columns
        :param list[str] old_cols: older columns

        :rtype: ColumnChanges
        """
        new_values = new_cols
        old_values = old_cols
        old_set = set(old_cols)
        new_set = set(new_cols)
        unchanged = old_set & new_set
        added = new_set - old_set
        removed = old_set - new_set
        return cls(new_values, old_values, unchanged, added, removed)


class DiffReader(object):
    """Interprets the changes between two commits.

    :var ColumnChanges column_changes: column changes
    :var ColumnChanges pk_changes: primary key changes
    :var RowIterator added_rows: iterator for added rows
    :var RowIterator removed_rows: iterator for removed rows
    :var ModifiedRowIterator modified_rows: iterator for modified rows
    :var TableProfileDiff data_profile: changes in data profile
    """

    column_changes: ColumnChanges
    pk_changes: ColumnChanges
    added_rows: RowIterator or None = None
    removed_rows: RowIterator or None = None
    modified_rows: ModifiedRowIterator or None = None
    data_profile: TableProfileDiff or None = None

    def __init__(self, repo: "repository.Repository", com_sum1: str, com_sum2: str, fetch_size: int = 100) -> None:
        """
        :param Repository repo: the repo handle
        :param str com_sum1: checksum of the first (newer) commit
        :param str com_sum2: checksum of the second (older) commit
        """
        dr = repo.diff(com_sum1, com_sum2)
        self.data_profile = dr.data_profile
        old_tbl = Table(columns=dr.old_columns, pk=dr.old_pk)
        new_tbl = Table(columns=dr.columns, pk=dr.pk)
        cd = ColDiff(old_tbl, new_tbl)
        self.column_changes = ColumnChanges.from_new_old_columns(
            dr.columns, dr.old_columns
        )
        self.pk_changes = ColumnChanges.from_new_old_columns(
            new_tbl.primary_key, old_tbl.primary_key
        )
        if dr.row_diff is not None and old_tbl.primary_key == new_tbl.primary_key:
            self.added_rows = RowIterator(
                repo=repo,
                tbl_sum=dr.table_sum,
                columns=new_tbl.columns,
                primary_key=new_tbl.primary_key,
                fetch_size=fetch_size
            )
            self.removed_rows = RowIterator(
                repo=repo,
                tbl_sum=dr.old_table_sum,
                columns=old_tbl.columns,
                primary_key=old_tbl.primary_key,
                fetch_size=fetch_size
            )
            self.modified_rows = ModifiedRowIterator(
                repo=repo,
                tbl_sum1=dr.table_sum,
                tbl_sum2=dr.old_table_sum,
                cd=cd,
                columns=[
                    col.name for col in cd.columns
                ],
                primary_key=new_tbl.primary_key,
                fetch_size=fetch_size
            )
            for rd in dr.row_diff:
                if rd.off1 is None:
                    self.removed_rows.add_offset(rd.off2)
                elif rd.off2 is None:
                    self.added_rows.add_offset(rd.off1)
                else:
                    self.modified_rows.add_offset(rd.off1, rd.off2)



# SPDX-License-Identifier: Apache-2.0
# Copyright © 2021 Wrangle Ltd

import typing
from unittest import TestCase

import attr

from wrgl.serialize import json_loads, json_dumps, field_transformer


@attr.s(auto_attribs=True, field_transformer=field_transformer(globals()))
class Person(object):
    name: str
    birth_date: str
    height: int
    scores: typing.List[int]


@attr.s(auto_attribs=True, field_transformer=field_transformer(globals()))
class Remote(object):
    url: str


@attr.s(auto_attribs=True, field_transformer=field_transformer(globals()))
class Branch(object):
    fetch: str


@attr.s(auto_attribs=True, field_transformer=field_transformer(globals()))
class Config(object):
    user: Person
    remotes: typing.List[Remote]
    branch: typing.Dict[str, Branch]


@attr.s(auto_attribs=True, field_transformer=field_transformer(globals()))
class Receive(object):
    deny_non_fast_forwards: bool
    deny_deletes: bool


class JSONTestCase(TestCase):
    def test_loads_simple(self):
        obj1 = Person(name='John Doe', birth_date='10/02/2000',
                      height=170, scores=[7, 8, 9])
        json_str = json_dumps(obj1)
        self.assertEqual(
            json_str, '{"name": "John Doe", "birthDate": "10/02/2000", "height": 170, "scores": [7, 8, 9]}'
        )
        obj2 = json_loads(json_str, Person)
        self.assertEqual(obj1, obj2)

    def test_loads_with_null(self):
        self.assertEqual(
            json_loads(
                '{"user": null, "remotes": null, "branch": null, "nonExistentKey": null}', Config
            ),
            Config()
        )

    def test_loads_nested(self):
        self.maxDiff = None
        obj1 = Config(
            user=Person(name='John Doe', birth_date='10/02/2000', height=170),
            remotes=[
                Remote(url='https://url1'),
                Remote(url='https://url2'),
            ],
            branch={
                'main': Branch(fetch='refs/heads/main'),
                'test': Branch(fetch='refs/heads/test')
            }
        )
        json_str = json_dumps(obj1, indent=4)
        self.assertEqual(
            json_str, '\n'.join([
                '{',
                '    "user": {',
                '        "name": "John Doe",',
                '        "birthDate": "10/02/2000",',
                '        "height": 170',
                '    },',
                '    "remotes": [',
                '        {',
                '            "url": "https://url1"',
                '        },',
                '        {',
                '            "url": "https://url2"',
                '        }',
                '    ],',
                '    "branch": {',
                '        "main": {',
                '            "fetch": "refs/heads/main"',
                '        },',
                '        "test": {',
                '            "fetch": "refs/heads/test"',
                '        }',
                '    }',
                '}',
            ])
        )
        obj2 = json_loads(json_str, Config)
        self.assertEqual(obj1, obj2)

    def test_loads_ignore_empty(self):
        obj = json_loads('{"denyNonFastForwards": true}', Receive)
        self.assertEqual(obj, Receive(deny_non_fast_forwards=True))
